cdef class High:
    cdef readonly int period
    cdef public list line
    cdef list window
    cdef list max
    cdef int current
    # 当前元素在整个序列中的位置
    def __init__(self, int period):
        self.period = period
        self.window = []
        self.line = []
        self.max = []
        self.current = 0
    def to_next(self, float data):
        cdef float res
        self.window.append(data)
        if self.current < self.period-1:
            # 淘汰上一个最大值
            while len(self.max) > 0 and self.window[self.max[-1] - self.current - 1] <= data:
                self.max.pop(-1)
            # 把当前元素添加上
            self.max.append(self.current)
            self.current += 1
            self.line.append(None)
            return None
        else:
            if len(self.max) >0 and self.max[0] <= self.current - self.period:
                self.max.pop(0)
            while len(self.max) > 0 and self.window[self.max[-1] - (self.current + 1 - self.period)] <= data:
                self.max.pop(-1)
            # 把当前元素添加上
            self.max.append(self.current)
            res = self.window[self.max[0] - (self.current + 1 - self.period)]
            self.window.pop(0)
            self.line.append(res)
            self.current += 1
            return res

export = High


from quotation.bar cimport Bar

cdef class Technique:
    def __cinit__(self, object indicator, float low, float high):
        self.indicator = indicator
        self.low = low
        self.high = high
    def check(self, Bar bar):
        cdef object v
        v = self.indicator.to_next(bar)
        return v is not None and self.low <= v <= self.high



from .high import High
from .low import Low


class Range:
    def __init__(self, period: int):
        self.period = period
        self.line = []
        self.high = High(period)
        self.low = Low(period)

    def to_next(self, high, low):
        _high = self.high.to_next(high)
        _low = self.low.to_next(low)
        _range = None
        if _high is not None and _low is not None:
            _range = _high - _low
        self.line.append(_range)
        return _range


export = Range



from quotation import LocalQuotation
from indicator import analyzer

import click
import tqdm
import traceback

import pandas as pd


@click.command()
@click.option('-i', '--input', help="Input File,which is a backtest result")
@click.option('-o', '--output', help="Output File,which should be a file with jpg/png ext.")
@click.option('-b', '--bundle', help='Quotation Data bundle.', default='ratio')
@click.option('-l', '--indicator_list', help='技术指标列表。用于说明如何生成技术指标。由XML格式所决定', default='indicators.xml')
def main(input, output, bundle, indicator_list):
    with pd.HDFStore(input) as bt_hdf:
        try:
            trades = bt_hdf.get('trades')
            params = bt_hdf.get('params')
            params = params.loc[0, :]
            instrument = params["instrument"]
            start_datetime = params["start_datetime"]
            end_datetime = params["end_datetime"]
            period = params["period"]

        except KeyError as k:
            print(f'Bad Input,occur key error:{k}')
            return
    try:
        ilist = analyzer.parseXMLToIndicators(indicator_list)
    except Exception:
        print(f'parse Indicator List XML :{indicator_list} Failed. Error is below ')
        traceback.print_exc()
        return
    try:
        quotation = LocalQuotation()
        quotation.load(start=start_datetime, end=end_datetime, bundle=bundle, instrument=instrument, frequency=period,
                       period_only=True)
    except Exception:
        print(
            f'Load quotation failed,time:[{start_datetime},{end_datetime}],'
            f'bundle={bundle},instrument={instrument},frequency={period}. Error is below')
        traceback.print_exc()
        return

    print('start to calculate')

    t = tqdm.tqdm(total=len(ilist), unit='indicator')

    def _update(t):
        return lambda: t.update(1)

    res = []

    for i in ilist:
        res1 = analyzer.analyze_instance(i=i, qt=quotation, trades=trades)
        res.append(res1)
        t.update(1)
    t.close()

    print('start to plot')

    t = tqdm.tqdm(total=len(res), unit='indicator')

    path = analyzer.plot_cumcurves(output, res, callback=_update(t))
    t.close()

    print(f'saved to {path}')


# qt.load(bundle = bundle)

if __name__ == '__main__':
    main()



"""[summary]
BacktestRunner的职责：
1. 根据数据源，构造中间商。
2. 实体化策略。
3. 在for-loop中回测策略
4. 【可选】分析策略，给出策略报表

我们希望这个能被更高级的（比如批量枚举的Runner给复用。
"""

from broker import LocalBroker
from runner import bt_impl
from runner.technique import Technique
import pandas as pd
import numpy as np

import math


def sharp_ratio(d):
    d = d.dropna()
    d = d.set_index('entry_datetime')
    profit = d['profit']
    profit = profit.resample('1D').sum()
    profit = profit[profit != 0]
    profit = np.asarray(profit)

    volatility = profit.std(ddof=1)

    if volatility != 0:
        avg_returns = profit.mean()
        return (avg_returns / volatility) * math.sqrt(252)
    else:
        return 0


def win_rate(d):
    d = d.loc[:, ['profit']]
    win = d[d['profit'] > 0].count()
    loss = d[d['profit'] < 0].count()

    return (win / (win + loss) * 100)['profit']


def plr(d):
    d = d.loc[:, ['profit']]
    win = d[d['profit'] > 0].mean()
    loss = d[d['profit'] < 0].mean().abs()
    return (win / loss)['profit']


def net_ret(d, initial_cash):
    ret_cash = d['profit'].sum()
    return ret_cash / initial_cash * 100


def max_drawdown(pnl):
    rec = list(pnl.items())
    last_max_time, last_max_value = rec[0]

    max_drawdown = 0
    max_drawdown_duration = np.timedelta64(0)

    for time, value in rec:
        if last_max_value < value:
            last_max_time = time
            last_max_value = value
        if last_max_value > value:
            dd = (last_max_value - value) / last_max_value
            if dd > max_drawdown:
                max_drawdown = dd
                ddd = time - last_max_time
                if ddd > max_drawdown_duration:
                    max_drawdown_duration = ddd
    return max_drawdown * 100, max_drawdown_duration


def conclude_from_trades(trade_df, initial_cash):
    df = trade_df.set_index('id')

    d = df.set_index('exit_datetime')
    d = d.dropna()
    pnl = d['profit'].cumsum() + initial_cash

    cost_functions = {}
    cost_functions['sharp_ratio'] = sharp_ratio(
        df)
    cost_functions['win_rate'] = win_rate(df)
    cost_functions['plr'] = plr(df)
    cost_functions['net_ret'] = net_ret(df, initial_cash)
    cost_functions['trade_count'] = df.shape[0]

    holding = (df['exit_datetime'] - df['entry_datetime'])

    cost_functions['max_holding'] = holding.max()
    cost_functions['min_holding'] = holding.min()
    cost_functions['avg_holding'] = holding.mean()
    cost_functions['std_holding'] = holding.std(ddof=0)
    cost_functions['max_drawdown'], cost_functions['max_drawdown_duration'] = max_drawdown(
        pnl)
    return df, pnl, cost_functions


class BackTestRunner:

    def __init__(self, str_module, ctx):
        """[summary]

        Arguments:
            str_module {[module]} -- [description]
            quotation {[Quotation]} -- [description]
            param_dict {[dict]} -- [description]
        """

        self.impl = bt_impl.BacktestRunnerImpl(str_module=str_module, ctx=ctx)

        self.str_module = str_module
        self.ctx = ctx

        self.trade_df = None
        self.cost_functions = None
        self.pnl = None
        self.techniques = []

    def attach_technique(self, t):
        self.impl.attach_technique(Technique(indicator=t["indicator"], low=t["low"], high=t["high"]))

    def attach_techniques(self, tl):
        techniques = []
        for t in tl:
            techniques.append(Technique(indicator=t["indicator"], low=t["low"], high=t["high"]))

        self.impl.attach_techniques(techniques)

    def run(self):
        self.impl.run()

    def conclude(self):

        broker = self.ctx.broker

        df = pd.DataFrame([trade.to_dict() for trade in broker.trades])
        if df.shape[0] == 0:
            return False
        df['submit_datetime'] = pd.to_datetime(df['submit_datetime'], unit='m')
        df['entry_datetime'] = pd.to_datetime(df['entry_datetime'], unit='m')
        df['exit_datetime'] = pd.to_datetime(df['exit_datetime'], unit='m')
        df = df[df['state'] != 3]
        self.trade_df, self.pnl, self.cost_functions = conclude_from_trades(df, broker.initial_cash)
        return True



cdef class EMA:
    cdef readonly int period
    cdef public list line
    cdef float k
    def __init__(self, int period):
        self.period = period
        self.line = []

        self.k = 2.0 / (period + 1)

    def to_next(self, float new_price):
        #574 ms ± 3.47 ms per loop (mean ± std. dev. of 7 runs, 250W loop each)
        cdef float result
        if len(self.line) == 0:
            self.line.append(new_price)
            return new_price
        else:
            result = new_price * self.k + self.line[-1] * (1 - self.k)
            self.line.append(result)
            return result

export = EMA



import numpy as np
cimport numpy as np

cdef class Dev:
    cdef np.ndarray window
    cdef int dof
    cdef int current
    cdef int window_size
    cdef readonly list line
    def __init__(self, int window_size, int dof):
        self.dof = dof
        self.window_size = window_size
        self.window = np.empty(dtype=np.float32, shape=(window_size,))
        self.current = 0
        self.line = []
    def to_next(self, float value):
        cdef float std
        if self.current < self.window_size:
            self.window[self.current] = value
            self.current += 1
            if self.current == self.window_size:
                std = np.std(self.window, ddof=self.dof)
                self.line.append(std)
                return std
            else:
                self.line.append(None)
                return None

        self.window[self.current % self.window_size] = value
        std = np.std(self.window, ddof=self.dof)
        self.line.append(std)
        return std

export = Dev



cimport numpy as np
import numpy as np

cdef class SimpleSlope:
    cdef np.ndarray window
    cdef int length
    cdef int current
    cdef list line
    def __init__(self, int length):
        self.length = length
        self.window = np.empty(dtype=np.float64, shape=(length,))
        self.current = 0
        self.line = []
    def to_next(self, np.float64_t data):
        cdef float slope
        cdef int index

        if self.current < self.length:
            self.window[self.current] = data
            self.current += 1

            if self.current != self.length:
                self.line.append(None)
                return None
            else:
                slope = (data - self.window[0]) / self.length
                self.line.append(slope)
                return slope

        index = self.current % self.length
        self.current +=1

        self.window[index] = data

        slope = (data - self.window[index - self.length+1]) / self.length

        self.line.append(slope)
        return slope



import numpy as np
cdef class Parameter:
    def __init__(self, float high, float low, float default_value):
        self.high = high
        self.low = low
        self.default_value = default_value

    def check_valid(self, float value):
        return self.low <= value < self.high

    def generate(self, int num):
        return np.linspace(self.low, self.high, num)




cdef class SMA:
    cdef readonly int period
    cdef public list line
    cdef list window
    cdef public float temp

    def __init__(self, int period):
        self.period = period
        self.line = []
        self.window = []
        self.temp = 0.0

    def to_next(self, float new_price):
        #574 ms ± 3.47 ms per loop (mean ± std. dev. of 7 runs, 250W loop each)
        cdef float last, result
        self.temp += new_price
        self.window.append(new_price)
        if len(self.window) < self.period:
            self.line.append(None)
            return None
        else:
            last = self.window.pop(0)
            result = self.temp / float(self.period)
            self.temp -= last
            self.line.append(result)
            return result

export = SMA


from trade.trade cimport Trade, Direction
from runner import Parameter

cimport numpy as np
from runner.context cimport Context
from quotation.bar cimport Bar

cpdef void init(Context ctx):
    # 这里可以指定交易的标的类型
    ctx.instrument = 'USDJPY'
    ctx.str_name = '波动回归策略'
    # 这里指定交易的频率（周期）
    ctx.frequency = '1D'

    #   不论是在优化还是回测，或者实盘运行中，init总共都只会被调用一次。
    #   在这里可以用Parameter类设置参数的最小值，最大值，默认值。
    #   1. 优化扫描时，将在最小和最大值之间按照调用程序设定的步长进行扫描。
    #   2. 回测和实际运行时， 如果调用者没有给出该参数的值，那么参数取默认值。

    ctx.parameters = {
        "k": Parameter(0.3, 0.1, 0.131),
        "t": Parameter(1.3, 1.0, 0.70),
        "s": Parameter(1.0, 0.5, 0.50),
        "quantity": Parameter(9999999999999, 0, 1),
    }

cpdef void on_new_bar(Context ctx, Bar bar):
    """[summary]
        In this function ,you can collect data by bars and do something.
    Arguments:
        ctx {[type]} -- [description]
        bar {[type]} -- [description]
    """

    pass

cpdef Trade on_new_bar_periodically(Context ctx, Bar last_bar, np.float64_t ask_open, np.float64_t bid_open):
    """[summary]
        在这个函数中，你可以根据上一个周期的数据以及当前周期开始的数据来判断你是否需要进行交易

        如果交易，按照下面的方式返回一个Trade即可
        如果不交易，那么返回None


    Arguments:
        ctx {[type]} -- [description]
        last_bar {[type]} -- [description]
        current_bar {[type]} -- [description]
    
    Returns:
        [type] -- [description]
    """
    cdef np.float64_t range_0
    cdef np.float64_t in_price, out_price, stop_price

    range_0 = last_bar.bid_open - last_bar.bid_low
    in_price = bid_open - range_0 * ctx.parameters_instance["k"]
    out_price = in_price + range_0 * ctx.parameters_instance["t"]
    stop_price = in_price - range_0 * ctx.parameters_instance["s"]

    # 下面以做多为例说明Trade的参数
    # 第一个参数是标的的类型，必须和你前面指定的一致
    # 第二个参数是入场价格，当你进行做多的时候，只有价格比这个低才会入场
    # 第三个参数是止损价格，当入场后价格跌破这个价格，那么会帮你止损出场（平仓）
    # 第四个是目标出场价格，当入场后价格超过这个价格的时候，会平仓止盈
    # 第五个参数是交易方向， BUY是做多， SELL 是做空

    # 如果一个周期内没有限价出场，也没有止损，那么在周期结束之前会以市价单的方式强行平仓
    return Trade(ctx.instrument, in_price, stop_price, out_price, ctx.parameters_instance["quantity"], Direction.BUY)



cdef class Low:
    cdef readonly int period
    cdef public list line
    cdef list window
    cdef list min

    cdef int current
    # 当前元素在整个序列中的位置

    def __init__(self, int period):
        self.period = period
        self.window = []
        self.line = []
        self.min = []
        self.current = 0

    def to_next(self, float data):
        cdef float res
        self.window.append(data)

        if self.current < self.period - 1:
            # 淘汰上一个最小值

            while len(self.min) > 0 and self.window[self.min[-1] - self.current - 1] >= data:
                self.min.pop(-1)

            # 把当前元素添加上
            self.min.append(self.current)
            self.current += 1
            self.line.append(None)
            return None
        else:
            if len(self.min)>0 and self.min[0] <= self.current - self.period:
                self.min.pop(0)
            while len(self.min) > 0 and self.window[self.min[-1] - (self.current + 1 - self.period)] >= data:
                self.min.pop(-1)
            # 把当前元素添加上
            self.min.append(self.current)
            res = self.window[self.min[0] - (self.current + 1 - self.period)]
            self.window.pop(0)
            self.line.append(res)
            self.current += 1
            return res

export = Low




cdef convert_bar_to_sgn(float open, float close):
    if open > close:
        return -1
    if open < close:
        return 1
    else:
        return 0

cdef class RSI:
    cdef readonly int period
    cdef public list line
    cdef float up
    cdef float down
    cdef list window

    def __init__(self, period):
        self.up = 0
        self.down = 0
        self.window = []
        self.line = []
        self.period = period
    def to_next(self, float new_open, float new_close):

        cdef float rs = 0.0, rsi = 0.0
        cdef float last = 0
        cdef float delta = 0

        delta = new_close - new_open

        if delta > 0:
            self.up += delta
        else:
            self.down += -delta

        self.window.append(delta)

        if len(self.window) < self.period:
            self.line.append(None)
            return None
        else:
            last = self.window.pop(0)
            rs = self.up + self.down
            if rs > 0:
                rsi = self.up / rs
            else:
                self.line.append(None)
                return None
            ## 只考虑多少两种情况
            if last > 0:
                self.up -= last
            else:
                self.down -= -last
            self.line.append(rsi)
            return rsi

export = RSI



from .bar import Bar
from .frequency import supported, deltas
import pandas as pd
import numpy as np
import array

from multiprocessing.shared_memory import SharedMemory

import config
import os


# 通过一组 API 来统一
# 这样不怕接口变动

def bundle_valid(bundle_path):
    return os.path.exists(bundle_path)


def ls_bundle(bundle_path):
    return os.listdir(bundle_path)


def get_freq_filename(bundle, instrument, frequency):
    return os.path.join(bundle, instrument, frequency + config.bundle_ext)


def frequency_supported(bundle, instrument, frequency):
    return os.path.exists(get_freq_filename(bundle, instrument, frequency))


def read_data(bundle, instrument, frequency):
    file = get_freq_filename(bundle, instrument, frequency)
    with pd.HDFStore(file) as hdf:
        data = hdf.get(config.bundle_key)
    return data


class LocalQuotation:
    def __init__(self):

        self.loaded = False

        self.len = 0
        self.len_p = 0

        self.data = None
        self.data_p = None

        # 用于存储共享内存的信息 （防止被释放）
        self.shm = None
        self.shm_p = None

        self.bar = []
        self.bar_p = []

        self.bundle = None

    def close_share(self):
        self.shm.close()
        self.shm_p.close()

        self.shm.unlink()
        self.shm_p.unlink()

    # 将数据分享
    def share_memory(self):
        """[summary]
            通过SharedMemory在进程间分享数据。
            注意：提供者的数据并不是在SharedMemory之上建立的
        Returns:
            [type] `dict` -- [description] 我们将共享内存块的名称，大小，类型以字典的方式返回，这样可以方便其他进程加载
        """
        assert self.loaded, 'Must be loaded!'

        self.shm = SharedMemory(create=True, size=self.data.nbytes)
        temp_array = np.ndarray(shape=self.data.shape, dtype=self.data.dtype, buffer=self.shm.buf)
        temp_array[:] = self.data[:]

        self.shm_p = SharedMemory(create=True, size=self.data_p.nbytes)
        temp_array = np.ndarray(shape=self.data_p.shape, dtype=self.data_p.dtype, buffer=self.shm_p.buf)
        temp_array[:] = self.data_p[:]

        return (self.shm.name, self.data.shape), (self.shm_p.name, self.data_p.shape)

        # 从共享内存加载数据

    def load_from_shared_memory(self, name_info):

        (name, shape), (name_p, shape_p) = name_info

        self.shm = SharedMemory(name=name)
        self.data = np.ndarray(shape=shape, dtype=np.float64, buffer=self.shm.buf)

        self.shm_p = SharedMemory(name=name_p)
        self.data_p = np.ndarray(shape=shape_p, dtype=np.float64, buffer=self.shm_p.buf)

        for data in self.data:
            self.bar.append(Bar(data))
        for data_p in self.data_p:
            self.bar_p.append(Bar(data_p))

        self.loaded = True

    # 从Bundle 文件中加载数据
    def load(self, instrument, frequency, start, end, bundle, period_only=False):
        """
            加载指定的行情数据包，你可以灵活地配置。
        :param instrument:  Instrument identifier
        :param frequency:  trade frequecny , you can lookup in quotation.frequency
        :param start: data start from
        :param end:  data end to
        :param bundle:  path to data bundle, like `data/bundles/ratio`
        :param period_only: whether only load periodical data
        :return:
        """
        assert bundle_valid(bundle), f' This Bundle({bundle}) is not valid!.'
        assert instrument in ls_bundle(bundle), f'This Instrument({instrument}) is not supported by Bundle {bundle}'
        assert frequency_supported(bundle, instrument,
                                   frequency), f'This Frequency({frequency}) is not supported,bundle:{bundle},instrument{instrument}'
        self.bundle = bundle

        if not period_only:
            # 读入分钟级别数据

            self.df = read_data(bundle, instrument, config.basic_freq)
            self.df = self.df[start:end]
            if 'Volume' not in self.df.columns:
                self.df.loc[:, 'Volume'] = 0
            self.df = self.df[
                ['AskOpen', 'AskHigh', 'AskLow', 'AskClose', 'BidOpen', 'BidHigh', 'BidLow', 'BidClose', 'Volume']]

            start_time = np.asarray(self.df.index, dtype=f'datetime64[{config.basic_unit}]')
            end_time = start_time + np.timedelta64(1, f'{config.basic_unit}')
            start_time = start_time.astype(np.float64).reshape((start_time.shape[0], 1,))
            end_time = end_time.astype(np.float64).reshape((start_time.shape[0], 1,))

            bar = np.asarray(self.df)

            self.data = np.concatenate((start_time, end_time, bar), axis=1)
            for data in self.data:
                self.bar.append(Bar(data))
        # 读入低频采样的数据

        self.df_p = read_data(bundle, instrument, frequency)
        self.df_p = self.df_p[start:end]
        if 'Volume' not in self.df_p.columns:
            self.df_p.loc[:, 'Volume'] = 0

        self.df_p = self.df_p[
            ['AskOpen', 'AskHigh', 'AskLow', 'AskClose', 'BidOpen', 'BidHigh', 'BidLow', 'BidClose', 'Volume']]

        start_time_p = np.asarray(self.df_p.index, dtype='datetime64[m]')
        end_time_p = start_time_p + deltas[frequency]
        start_time_p = start_time_p.astype(np.float64).reshape((start_time_p.shape[0], 1,))
        end_time_p = end_time_p.astype(np.float64).reshape((start_time_p.shape[0], 1,))

        bar_p = np.asarray(self.df_p)

        self.data_p = np.concatenate((start_time_p, end_time_p, bar_p), axis=1)

        for data_p in self.data_p:
            self.bar_p.append(Bar(data_p))
        self.loaded = True



from .base import Broker
from trade.trade cimport State, Trade, OrderType, Direction
from trade.trade import Trade
from quotation.bar cimport Bar
from quotation.bar import Bar
import numpy as np
from broker.execute_local import execute_limit_order_s, execute_limit_order, execute_market, execute_stop_order_s, \
    execute_stop_order,execute_els_order
from util.forex import point_rate

# 根据交易细节断定能不能成交
cdef inline float calculate_profit(float exit_price, float entry_price, float quantity, Direction direction,
                                   str instrument):
    return (exit_price - entry_price) * quantity * point_rate[
        instrument] * direction

cdef class LocalBroker:
    def __init__(self, float commission_rate=0.0, float initial_cash=1000.0):
        """[summary]
            This local broker only supports single trade mode.
            Every method of the local broker should be called only by the runner, which means you should
            ensure everything works as the real trading envrionment。
            So this api is not useful for user level of `newtrader`.

            Every action,such as submit/exit trades should be provided with the accurate context data including quotation and datetime.

            It‘s not neccessary for it to be provided with the quotation in the initialization funciton.

            In fact, when trading online, quotation is provided by the broker itself, which means it;s not our bussiness.

        
        Keyword Arguments:
            commission_rate {float} -- [description] (default: {0.0})
            initial_cash {float} -- [description] (default: {1000.0})
        """
        self.trade_id = 0
        self.trades = []

        self.commission_rate = commission_rate

        self.initial_cash = initial_cash
        self.cash = initial_cash

        self.current_trade = None

    def submit_trade(self, Trade trade, object datetime):
        """[summary]
            我们把这个交易提交到做市商
        Arguments:
            trade {[Trade]} -- [description] 要提交的交易
        """

        trade.accept(datetime, self.trade_id)
        self.trade_id += 1

        self.trades.append(trade)
        self.current_trade = trade
    def close_trade(self, Bar bar):
        """[summary]
            Close the current trade.

            Exit market when entered, else cancel it.
        Arguments:
            last_bar {[type]} -- [description]
            bar {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        cdef float profit

        if self.current_trade is None:
            return False

        assert self.current_trade.state in [State.INITIAL,
                                            State.ENTRIED], f'Trade [{self.current_trade.id}] cannot be closed.'

        if self.current_trade.state == State.INITIAL:
            self.current_trade.cancel()
        else:
            price = execute_market(bar=bar, direction=self.current_trade.direction * (-1))
            profit = calculate_profit(exit_price=price, entry_price=self.current_trade.entryPrice,
                                      direction=self.current_trade.direction,
                                      quantity=self.current_trade.quantity,
                                      instrument=self.current_trade.instrument)
            self.current_trade.exit(exit_price=price, profit=profit, exit_type=OrderType.Market, at=bar.start)
            self.cash += self.current_trade.profit

        self.current_trade = None

        return True

    # 每个周期开始时第一个Bar的处理
    def on_bar_p_start(self, Bar bar):
        cdef float price
        price = -1
        if self.current_trade is not None:
            if self.current_trade.entry_type == OrderType.Limit:
                price = execute_limit_order_s(bar, self.current_trade.direction, self.current_trade.entryPrice)
            if self.current_trade.entry_type == OrderType.Stop:
                price = execute_stop_order_s(bar, self.current_trade.direction, self.current_trade.entryPrice)
        if price > 0:
            self.current_trade.entry(entryPrice=price, at=bar.start)
    def on_bar(self, Bar last_bar, Bar bar):
        cdef float profit

        if self.current_trade is None:
            return

        if self.current_trade.state == State.INITIAL:
            if self.current_trade.entry_type == OrderType.Limit:
                price = execute_limit_order(last_bar, bar, self.current_trade.direction, self.current_trade.entryPrice)
            if self.current_trade.entry_type == OrderType.Stop:
                price = execute_stop_order(last_bar, bar, self.current_trade.direction, self.current_trade.entryPrice)
            if price > 0:
                self.current_trade.entry(entryPrice=price, at=bar.start)

        if self.current_trade.state == State.ENTRIED:
            # 出场的时候和入场的方向相反
            price, type = execute_els_order(last_bar=last_bar, bar=bar, direction=self.current_trade.direction * (-1),
                                            stop_price=self.current_trade.stopLossPrice,
                                            limit_price=self.current_trade.profitTargetPrice)
            if price > 0:
                profit = calculate_profit(exit_price=price, entry_price=self.current_trade.entryPrice,
                                          direction=self.current_trade.direction,
                                          quantity=self.current_trade.quantity,
                                          instrument=self.current_trade.instrument)
                self.current_trade.exit(exit_price=price, profit=profit, exit_type=type, at=bar.start)
                self.current_trade = None



from quotation.bar cimport Bar
from trade.trade cimport Direction, OrderType

cpdef float execute_market(Bar bar, Direction direction):
    if direction == Direction.BUY:
        return bar.ask_close
    else:
        return bar.bid_close

cpdef float execute_limit_order(Bar last_bar, Bar bar, Direction direction, float price):
    if direction == Direction.BUY:
        if bar.ask_high < price:
            return bar.ask_high
        if last_bar.ask_close > price >= bar.ask_low:
            return price
        else:
            return -1.0
    if direction == Direction.SELL:
        if bar.bid_low > price:
            return bar.bid_low
        elif last_bar.bid_close < price <= bar.bid_high:
            return price
        else:
            return -1.0

cpdef float execute_stop_order(Bar last_bar, Bar bar, Direction direction, float price):
    if direction == Direction.BUY:
        if bar.ask_low > price:
            return bar.ask_low
        elif last_bar.ask_close < price <= bar.ask_high:
            return price
        else:
            return -1.0
    if direction == Direction.SELL:
        if bar.bid_high < price:
            return bar.bid_high
        elif last_bar.bid_close > price >= bar.bid_low:
            return price
        else:
            return -1.0
cpdef float execute_limit_order_s(Bar bar, Direction direction, float price):
    if direction == Direction.BUY:
        if bar.ask_close <= price:
            return price
        else:
            return -1.0
    if direction == Direction.SELL:
        if bar.bid_close >= price:
            return price
        else:
            return -1.0
cpdef float execute_stop_order_s(Bar bar, Direction direction, float price):
    if direction == Direction.BUY:
        if bar.ask_close >= price:
            return price
        else:
            return -1.0
    if direction == Direction.SELL:
        if bar.bid_close <= price:
            return price
        else:
            return -1.0

cpdef tuple execute_els_order(Bar last_bar, Bar bar, Direction direction, float stop_price, float limit_price):
    cdef float price
    price = execute_limit_order(last_bar, bar, direction, limit_price)
    if price > 0:
        return price, OrderType.Limit
    price = execute_stop_order(last_bar, bar, direction, stop_price)
    if price > 0:
        return price, OrderType.Stop
    return -1.0,OrderType.INVALID



cimport numpy as np
cpdef enum Direction:
    BUY = 1
    SELL = -1

cpdef enum State:
    INITIAL = 0
    ENTRIED = 1
    EXITED = 2
    CANCELLED = 3

cpdef enum OrderType:
    Limit = 0
    Market = 1
    Stop = 2
    INVALID = -1

cdef class Trade:
    cdef readonly  int id
    cdef readonly  str instrument
    cdef readonly  np.float64_t entryPrice
    cdef readonly  np.float64_t stopLossPrice
    cdef readonly  np.float64_t profitTargetPrice
    cdef readonly  int quantity
    cdef readonly  object submit_datetime
    cdef readonly  object exit_datetime
    cdef readonly object entry_datetime
    cdef readonly np.float64_t exit_price
    cdef readonly  Direction direction
    cdef readonly  np.float64_t commission_rate
    cdef readonly  np.float64_t commission
    cdef readonly  np.float64_t profit # 设定后会自动扣除手续费
    cdef readonly OrderType entry_type
    cdef readonly  OrderType exit_type
    cdef readonly  State state




from trade.trade cimport State, Trade

cdef class LocalBroker:
    cdef readonly list trades
    cdef int trade_id
    cdef readonly float commission_rate
    cdef readonly float initial_cash
    cdef public float cash
    cdef Trade current_trade




import numpy as np

# cdef class Parameter:
#     def __init__(self, float high, float low, float default):
#         self.high = high
#         self.low = low
#         self.default = default
#         assert self.low <= self.default < self.high, f'Should low= {low} <=default= {default} < high= {high}'
#
#     def check_valid(self, float value):
#         return self.low <= value < self.high
#
#     def generate(self, int num):
#         return np.linspace(self.low, self.high, num)


cdef class Context:
    def __cinit__(self,object quotation=None,str instrument="",float initial_cash=0.0,float commission_rate=0.0,
                    object start_date=None,object end_date=None,str frequency='1D',object broker = None
                 ):
        """[summary]
            Initialize the context and run the result.

        """
        self.instrument = instrument
        self.quotation = quotation

        self.broker = broker

        self.start_date = start_date
        self.end_date = end_date

        self.parameters = {}
        self.parameters_instance = {}

        self.initial_cash = initial_cash
        self.commission_rate = commission_rate

        self.str_name = 'Unamed Stragety'

        self.frequency = frequency



cimport numpy as np
cimport quotation.bar
import quotation.bar

cdef class Trade:
    """
        对于 Trade 这个类，它包含了交易的一些信息
        但是我们的设计并不希望他包含执行的逻辑。
        执行的逻辑应该由具体的Broker的配套给出
    """

    def __init__(self, str instrument, np.float64_t entryPrice, np.float64_t stopLossPrice,
                 np.float64_t profitTargetPrice, int quantity,
                 Direction direction, OrderType entry_type):
        # assert type(direction) is Direction
        # 
        # assert (
        #                stopLossPrice - entryPrice) * direction.value <= 0, "Invalid Stop Loss Price:{},with Entry Price:{},direction:{}".format(
        #     stopLossPrice, stopLossPrice, "Buy" if (direction == 1) else "Sell")
        # assert (profitTargetPrice - entryPrice) * direction.value >= 0, "Invalid Targe price"

        self.id = 0
        self.instrument = instrument

        self.entryPrice = round(entryPrice, 5)
        self.stopLossPrice = round(stopLossPrice, 5)
        self.profitTargetPrice = round(profitTargetPrice, 5)
        self.quantity = quantity

        # 关于时间： 我们用最小频率的开始的那个时刻做记录
        self.submit_datetime = None
        self.entry_datetime = None
        self.exit_datetime = None
        self.exit_price = 0

        self.direction = direction

        self.commission_rate = 0
        self.commission = 0

        self.profit = 0.0

        self.entry_type = entry_type
        self.exit_type = OrderType.INVALID
        self.state = State.INITIAL

    def to_dict(self):
        return {
            "id": self.id,
            "instrument": self.instrument,
            "direction": self.direction,
            "entryPrice": self.entryPrice,
            "stopLossPrice": self.stopLossPrice,
            "profitTargetPrice": self.profitTargetPrice,
            "quantity": self.quantity,
            "submit_datetime": self.submit_datetime,
            "entry_datetime": self.entry_datetime,
            "exit_datetime": self.exit_datetime,
            "entry_type": self.entry_type,
            "exit_price": self.exit_price,
            "exit_type": self.exit_type,
            "commission": self.commission,
            "state": self.state,
            "profit": self.profit,
        }

    def set_commission_rate(self, float rate):
        self.commission_rate = rate
        self.commission = rate * self.quantity

    def accept(self, object dt, int ID):
        """[summary]
            Accept the trade to the broker.
            When it's accepted , an ID should be provided.
        Arguments:
            dt {[type]} -- [description]
            ID {[type]} -- [description]
        """
        self.id = ID
        self.submit_datetime = dt

    def cancel(self):
        """[summary]
            If the trade is just initialized,it can be cancelled.

            If cancelled retrun `False`.
            else `True`
        Arguments:
            dt {[type]} -- [description]
        
        Returns:
            [type] -- [description]
        """
        assert self.state == State.INITIAL, f"The trade cannot be exited for that the state for trade " \
                                            f"{self.id}, start at {self.submit_datetime} is {self.state},entering at f{self.entry_datetime}."
        self.state = State.CANCELLED

    def entry(self, float entryPrice, object at):
        self.entryPrice = entryPrice
        self.entry_datetime = at
        self.state = State.ENTRIED
    def exit(self, OrderType exit_type, float exit_price, float profit, object at):
        """[summary]
            由Broker调用，由行情数据判断如何退出市场
        Arguments:
            last_bar {[type]} -- [description]
            bar {[type]} -- [description]

        Returns:
            [type] -- [description]
        """

        self.exit_price = exit_price
        self.exit_datetime = at
        self.profit = profit - self.commission
        self.exit_type = exit_type
        self.state = State.EXITED



from numpy cimport float64_t, ndarray
from numpy import ndarray
import ctypes
from libc.stdlib cimport malloc, free

cdef class Bar:
    def __cinit__(self, np.ndarray[np.float64_t] array):
        self.ask_open= array[2]
        self.ask_high= array[3]
        self.ask_low= array[4]
        self.ask_close= array[5]
        self.bid_open= array[6]
        self.bid_high= array[7]
        self.bid_low= array[8]
        self.bid_close= array[9]
        self.volume= array[10]
        self.start= array[0]
        self.end= array[1]



from runner import BackTestRunner, Context
from multiprocessing import Process, Manager
from quotation import LocalQuotation

import traceback
import os
import queue


def work(str_module_name, commission_rate, frequency, shm_names, p, q):
    str_module = __import__(str_module_name)

    qt = LocalQuotation()
    qt.load_from_shared_memory(shm_names)

    temp_ctx = Context()
    temp_ctx.quotation = qt
    temp_ctx.commission_rate = commission_rate
    temp_ctx.frequency = frequency

    while True:
        try:
            param = p.get_nowait()
            temp_ctx.parameters_instance = param
            temp_ctx.instrument = param["instrument"]
            temp_ctx.initial_cash = param['initial_cash']
            bt = BackTestRunner(str_module, temp_ctx)
            bt.run()
            bt.conclude()

            q.put((param, bt.cost_functions))
        except queue.Empty as e:
            return
        except Exception as e:
            traceback.print_exc()


def run_scan(str_module_name, ctx, params, process_count=None):
    results = []
    processes = []
    manager = Manager()

    q = manager.Queue()
    p = manager.Queue()

    if process_count is None:
        process_count = os.cpu_count()

    try:

        quotation_shm_name = ctx.quotation.share_memory()

        # 发送参数
        for param in params:
            p.put(param)

        for i in range(process_count):
            processes.append(Process(group=None, name=f'newtrade_scanner_{i}', target=work,
                                     args=(str_module_name, ctx.commission_rate,
                                           ctx.frequency, quotation_shm_name, p, q)))

        for process in processes:
            process.start()

        from tqdm import tqdm
        t = tqdm(total=len(params), unit='task', smoothing=False)
        t.update(0)
        for i in range(len(params)):
            params, cost_functions = q.get(timeout=None)
            results.append(dict(params, **cost_functions))
            t.update()
        t.close()

        # 等待结束
        for process in processes:
            process.join(timeout=None)

    except Exception as e:
        t.close()
        traceback.print_exc()

    return results



from runner.context cimport Context
from runner.context import Context
from broker.local import LocalBroker
from broker.local cimport LocalBroker
from trade.trade cimport Trade
from quotation.bar cimport Bar
from quotation.bar import Bar
from runner.technique cimport Technique
from runner.technique import Technique

cdef class BacktestRunnerImpl:
    cdef Context ctx
    cdef object str_module
    cdef LocalBroker broker
    cdef list techniques
    def __cinit__(self, object str_module, Context ctx):
        self.str_module = str_module
        self.ctx = ctx

        self.broker = LocalBroker(ctx.commission_rate, ctx.initial_cash)

        self.ctx.broker = self.broker

        self.techniques = []

    def attach_technique(self, Technique technique):
        self.techniques.append(technique)
    def attach_techniques(self, list techniques):
        self.techniques += techniques

    def check_technique(self, Bar bar):
        cdef Technique t
        for t in self.techniques:
            if not t.check(bar):
                return False
        return True

    cpdef public void run(self):
        cdef Bar bar,last_bar,bar_p,last_bar_p
        cdef Trade trade
        assert self.ctx.quotation.loaded, "Data Must be Loaded"

        last_bar_p = None

        it = iter(self.ctx.quotation.bar)
        bar =(next(it))
        last_bar = None
        try:
            for bar_p in self.ctx.quotation.bar_p:
                trade = None


                # if last_bar_p is not None and self.check_technique(last_bar_p):
                if last_bar_p is not None and self.check_technique(last_bar_p):
                    trade = self.str_module.on_new_bar_periodically(
                        self.ctx, last_bar_p, ask_open=bar_p.ask_open, bid_open=bar_p.bid_open)

                if trade is not None:
                        # 提交订单以开始的时间为准
                        self.broker.submit_trade(trade, bar_p.start)

                # 单是下在这个周期开始的时间的，那么理论上这个周期内之后的所有分钟数据都可以使用
                # 但是需要注意一下；
                # 对于这个周期内的第一个分钟，只有在整个Bar 都满足成交条件才能成交
                # 对于后面的Bar，则是仅仅需要最高价（或者最低价满足成交条件）
                # 并且前面一个Bar没有达到成交条件。



                # 上一个Bar必须被清空。。


                # 如果分钟远远没有达到这个周期的时间，就过一下
                while bar.start <bar_p.start:
                    bar = next(it)

                last_bar = None

                self.broker.on_bar_p_start(bar)

                while bar_p.start <= bar.start and bar.end <=bar_p.end:
                    self.broker.on_bar(last_bar, bar)

                    last_bar = bar
                    bar = next(it)
                # when the loop is end,the minute is going to be in the next period, or the quotation length.



                # anyway,we close the trade. We don't care about the result.
                # When close the trade, we use the data of the last bar , because the trade should be closed before the
                # period is end.
                self.broker.close_trade(last_bar)

                if self.broker.cash <= 0:
                    # You are broken now.
                    break

                last_bar_p = bar_p
        except StopIteration:
            return



import glob
import importlib
import sys
from os.path import dirname, join, basename, isfile

modules1 = glob.glob(join(dirname(__file__), "*.py"))
modules2 = glob.glob(join(dirname(__file__), "*.pyx"))
__all__ = [basename(f)[:-3] for f in modules1 if isfile(f) and not f.endswith('__init__.py')]
__all__ += [basename(f)[:-4] for f in modules2 if isfile(f)]

classes = {}

for indicator in __all__:
    module = importlib.import_module('.' + indicator, 'indicator.indicators')
    if hasattr(module, 'export'):
        iClass = module.export
        setattr(sys.modules['indicator'], indicator.upper(), iClass)
        classes[indicator.upper()] = iClass



import matplotlib.pyplot as plt
import os
import datetime
import matplotlib.ticker as tk
import matplotlib.dates as mdates  # 處理日期
from matplotlib.backends.backend_pdf import PdfPages


def trim_axs(axs, N):
    """little helper to massage the axs list to have correct length..."""
    axs = axs.flat
    for ax in axs[N:]:
        ax.remove()
    return axs[:N]


def cb():
    pass


def plot_cumcurve(axs, data, callback):
    (name, start_v, end_v, filtered_pnl, cs, indicator_df) = data
    ax1, ax2, ax3 = axs

    ax1.plot(cs.index, cs['profit'], 'o-', color='red', markersize=2, )
    ax1.set_xlabel(name)
    ax1.set_ylabel('Cum Return')
    ax1.set_title(f'{round(start_v,8)}<{name}<{round(end_v,8)}')
    ax1.grid(True)
    ax1.xaxis.set_major_locator(tk.MaxNLocator())
    ax1.axvline(start_v, color='blue', label=start_v)
    ax1.axvline(end_v, color='green', label=end_v)

    ax2.plot(indicator_df.index, indicator_df[name], 'o-', markersize=2, color='red')
    ax2.fmt_xdata = mdates.DateFormatter('%Y-%m')
    ax2.grid(True)
    ax2.set_title('Indicator Series')
    ax2.set_xlabel('Date')
    ax2.set_ylabel(name)
    plt.setp(ax2.get_xticklabels(), rotation=30, ha='right')

    ax3.plot(filtered_pnl.index, filtered_pnl, 'o-', markersize=2, color='red')
    # ax3.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m'))
    # ax3.xaxis.gcf().autofmt_xdate()
    ax3.fmt_xdata = mdates.DateFormatter('%Y-%m')
    ax3.set_title('Filtered P&L')
    ax3.set_xlabel('Date')
    ax3.set_ylabel('Returns')
    plt.setp(ax3.get_xticklabels(), rotation=30, ha='right')
    ax3.grid(True)
    callback()


def plot_page(pp, page_data, iheight, iwidth, dpi, callback):
    count = len(page_data)
    fig: plt.Figure = plt.figure(figsize=(iwidth * 3 + 5, iheight * count + 5), dpi=dpi)
    axss = fig.subplots(nrows=count, ncols=3)
    for ax, data in zip(axss, page_data):
        plot_cumcurve(ax, data, callback)
    fig.tight_layout()
    fig.savefig(pp, format='pdf')
    plt.close(fig)


def plot_cumcurves(save_path, data, iheight=5, iwidth=5, dpi=32, callback=cb, maxrow=5):
    if save_path is None:
        if not os.path.exists('images'):
            os.makedirs('images')

        save_path = os.path.join('images', datetime.datetime.now().isoformat() + '.pdf')

    assert save_path.endswith('.pdf'), 'must be saved in PDF!'

    pp = PdfPages(save_path)

    # 对数据进行一个分割
    splited = [data[i:i + maxrow] for i in range(0, len(data), maxrow)]

    for page in splited:
        plot_page(pp=pp, page_data=page, iheight=iheight, iwidth=iwidth, dpi=dpi, callback=callback)

    pp.close()

    return save_path
